import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.net.Socket;
import java.util.Iterator;
import java.util.Map.Entry;

/** 
 *  UsaProxy - HTTP proxy for tracking and logging of user interactions on websites
    Copyright (C) 2006  Monika Wnuk - Media Informatics Group at the University of Munich
    <br><br>
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    <br><br>
 *  Class SocketData provides methods for client and server socket
 *  connections such as the management of input and output streams.
 *  It also reads in the request/response headers and manages them 
 *  in an OrderedHashtable.
 */

public class SocketData {
	
	private Socket 				socket;
	private OrderedHashtable 	headers;
    private InputStream 		in; 
    private OutputStream 		out;
	
    /** Constructor, creates a SocketData class for a given Socket.
     * @param socket is the Socket
     */
	public SocketData(Socket socket) {
		this.socket 	= socket;
		this.headers	= new OrderedHashtable();
	}
	
	/**
     *  Sets the InputStream of the respective Socket.
     */
	public void getInputStream() throws IOException {
		this.in = socket.getInputStream();
	}
	
	/**
     *  Sets the OutputStream of the respective Socket.
     */
	public void getOutputStream() throws IOException {
		this.out = socket.getOutputStream();
	}
	
	/**
     *  Closes the InputStream of the respective Socket.
     */
	public void closeInputStream() throws IOException {
		if (this.in != null) this.in.close();
	}
	
	/**
     *  Closes the OutputStream of the respective Socket.
     */
	public void closeOutputStream() throws IOException {
		if (this.out != null) this.out.close();
	}
	
	/**
     *  Closes the Socket.
     */
	public void closeSocket() throws IOException {
		if (this.socket != null) this.socket.close();
	}
    
    /**
     * Returns the Socket
	 * @return the socket.
	 */
    public Socket getSocket() {
		return socket;
	}
    
	/**
	 * Sets the Socket
	 * @param socket is the socket to set.
	 */
	public void setSocket(Socket socket) {
		this.socket = socket;
	}

	/**
     *  Sets the InputStream of the respective Socket.
     *  @param in is the InputStream to set
     */
	public void setIn(InputStream in) {
		this.in = in;
	}

	/**
     *  Sets the OutputStream of the respective Socket.
     *  @param out is the OutputStreams to set
     */
	public void setOut(OutputStream out) {
		this.out = out;
	}
	
	/**
	 * Returns the InputStream of the respective Socket
	 * @return the InputStream.
	 */
	public InputStream getIn() {
		return in;
	}
	
	/**
	 * Returns the OutputStream of the respective Socket
	 * @return the OutputStream.
	 */
	public OutputStream getOut() {
		return out;
	}

	/**
	 * Returns the OrderedHashtable of the respective Socket
	 * @return the OrderedHashtable.
	 */
	public OrderedHashtable getHeaders() {
		return headers;
	}

	/**
	 * Sets the OrderedHashtable of the respective Socket
	 * @param headers is the OrderedHashtable to set.
	 */
	public void setHeaders(OrderedHashtable headers) {
		this.headers = headers;
	}
	
	/**
     *  Reads in the HTTP request/response headers and stores them
     *  in the OrderedHashtable headers.
     *  Return the remaining input stream for further processing.
     *  @param in is the InputStream the headers are read from.
     *  @return the remaining InputStream
     */
	public InputStream readHeaders(InputStream in) throws IOException {
		byte b 			= 0;
        int len 		= 0;
        int colonIndex	= -1;
        String name, value, tmpString;
        
        if (in == null)
			in = new BufferedInputStream(in);

        if (getHeaders() == null) 
        	setHeaders(new OrderedHashtable());
            
        
        /** Logic to read HTTP request headers */
        boolean readTooMuch = false;

        for (ByteArrayOutputStream buf = new ByteArrayOutputStream(4097); ;) {
            if (!readTooMuch) {
                b = (byte) in.read();
            }
            if (b == -1) {
                break;
            }
            readTooMuch = false;
            if ((b != '\r') && (b != '\n')) {
                if ((b == ':') && (colonIndex == -1)) {
                    colonIndex = len;
                }
                len++;
                buf.write(b);
            } else if (b == '\r') {
                continue;
            } else {    // b== '\n'
                if (len == 0) {
                    break;
                }
                b = (byte) in.read();
                readTooMuch = true;

                /** A space or tab at the beginning of a line means the header continues */
                if ((b == ' ') || (b == '\t')) {
                    continue;
                }
                buf.close();
                byte[] hdata = buf.toByteArray();
                buf.reset();
                if (colonIndex != -1) {  // if ":" exists in header
                	tmpString = new String(hdata, 0, len, HTTPData.HEADER_DEFAULT_CHAR_ENCODING);
                	if (HTTPData.isRequestLine(tmpString)) {
                		name = tmpString;
                		value = "";
                	} else {
                		name 	= new String(hdata, 0, colonIndex, HTTPData.HEADER_DEFAULT_CHAR_ENCODING);
                		value 	= new String(hdata, colonIndex + 2, len - 2 - colonIndex,
                                    HTTPData.HEADER_DEFAULT_CHAR_ENCODING);
                	}
                    colonIndex = -1;
                } else {
                    name = new String(hdata, 0, len,
                                    HTTPData.HEADER_DEFAULT_CHAR_ENCODING);
                    value = "";
                }
                
                if (value=="") getHeaders().add(0, name, value);
                else getHeaders().put(name, value);
                len = 0;
            }
        }
        
        return in;
	}
	
	/**
     *  Prints the entries of the OrderedHashtable headers.
     */
	public void printHeaders() {
        StringBuffer headers = new StringBuffer();
		
		for (Iterator e = this.getHeaders().entrySet().iterator(); e.hasNext();) {
	
	        Entry me 	= (Entry) e.next();
	        String key 	= (String) me.getKey();
	        if (null == key) continue;
	        if (HTTPData.isStatusLine(key.trim()))
	        	headers.insert(0,key + "\r\n");
	        else {
	        	String value2 = (String)me.getValue();
	        	headers.append(key).append(": ").append(value2).append("\r\n");
	        }
	    }
		headers.append("\r\n");
		System.err.println("Response headers: " + headers);
    }
	
	/**
     *  Sends error message to the client.
     *  @param err is the Exception that occurred.
     *  @param get is the URL the client has requested.
     *  @param host is the host the URL is associated with.
     */
	public void sendErrorMessage(Exception err, String get, String host) {
    	
    	/** Send error message to client */
		PrintWriter out = new PrintWriter(new OutputStreamWriter(this.getOut()));

		out.println("HTTP/1.0 200 ");            /** version and status code */
		out.println("Content-Type: text/plain"); /** data will be sent as text */
		out.println();                           /** blank line */

		out.println("GET/POST:  " + get);
		out.println("HOST: " + host);
		out.println("Error!");
		out.println(err);
		out.flush();
    }
}
